#include <stdio.h>
#include <sys/socket.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/select.h>
#include "sendto_.c"

#define PORT 2222
#define BLOCKSIZE 128
#define WINDOWSIZE 4

#define TIMEOUT 5

typedef struct PACKET {
	unsigned char number;
	unsigned int size;
	char data[BLOCKSIZE];
} packet;

typedef struct CLIENT_INTERNAL {
	packet data;
	unsigned int  timeout;
	unsigned int  valid;  /* Test platnosti dat v okne pro vysilani */
} client_internal;

typedef struct AKNOWLEDGE {
	unsigned char number;
	enum { DATA, END } type;
} acknowledge;

/*********************************************************************/

int main(int argc, char *argv[]){
	int sock, fd;
	struct sockaddr_in adresa;
	client_internal *window = (client_internal *) malloc(WINDOWSIZE * sizeof(client_internal));
	int i;	
	unsigned char zacatek = 0, packet_number, end_flag = 0, finish;
	fd_set fd_test;
	struct timeval timeout;
	acknowledge ack;	
	
	sock = socket(AF_INET, SOCK_DGRAM, 0);

	memset(&adresa, 0, sizeof(adresa));

	adresa.sin_port = htons(PORT);
	adresa.sin_family = AF_INET;
	adresa.sin_addr.s_addr = inet_addr("127.0.0.1");

	if(connect(sock, (struct sockaddr *) &adresa, sizeof(adresa)) < 0){
		fprintf(stderr, "connect: %s\n", strerror(errno));
		exit(-1);	
	}

	/* Posilani souboru */
	if((fd = open(argv[1], O_RDONLY)) == -1){
		fprintf(stderr, "open: %s\n", strerror(errno));
		exit(-1);	
	};

	/* Pripraveni okna */
	for(i = 0; i < WINDOWSIZE; i++){
		window[i].valid = !end_flag;
		if(!end_flag){
			window[i].data.size = read(fd, window[i].data.data, BLOCKSIZE);
			window[i].data.number = i;
			window[i].timeout = 0;

			if(window[i].data.size != BLOCKSIZE){
				end_flag = 1;
			}

			printf("Nacitam %d bytu\n", window[i].data.size);
		}
	}
	packet_number = WINDOWSIZE;
	
	/* Hlavni cyklus */
	for(;;){
		finish = 0;

		for(i = 0; i < WINDOWSIZE; i++){
			if(window[i].valid){
				finish ++;
				if(window[i].timeout == 0){
/*					usleep(PAUSE);*/
					sendto_(sock, &(window[i].data), sizeof(packet), 0, (struct sockaddr *) &adresa, sizeof(adresa));	
					/* nastavit timeout */
					window[i].timeout = TIMEOUT;
				}
				else {
					window[i].timeout-- ;
				}
			}

		}
		if(finish == 0){
			 /* test na uplny konec */
//			close(sock);
			close(fd);
//			exit(0);
			break;
		}
		
		FD_ZERO(&fd_test);		
		FD_SET(sock, &fd_test);
	
		timeout.tv_sec = 0;
		timeout.tv_usec = 500000;

		/* Cekani na ACK */
		if( select(sock + 1, &fd_test, 0, 0, &timeout) == 1 ){
			recvfrom(sock, &ack, sizeof(ack), 0, 0, 0);	

			printf("ACK: %d\n", ack.number);
			
			if(ack.number >= zacatek && ack.type == DATA){
			/* Posunuti okna na zacatek ACK */
				unsigned char tmp = ack.number - zacatek + 1;
				if(tmp > WINDOWSIZE){
					/* Nedelej nic */
				}
				else{
					memmove(window, window + tmp, sizeof(client_internal)*(WINDOWSIZE - tmp));
					zacatek = ack.number + 1;
					
					/* Donacteni */
					for(i = 0; i < tmp; i++){
						int index = i + WINDOWSIZE - tmp;
						window[index].valid = !end_flag;
						if(!end_flag){
							window[index].data.size = read(fd, window[index].data.data, BLOCKSIZE);
							window[index].data.number = packet_number++;
							window[index].timeout = 0;
							
							printf("Nacitam packe %d. -- %d bytu\n", window[index].data.number, window[index].data.size);
			
							if(window[index].data.size != BLOCKSIZE){
								end_flag = 1;
							}
						}
					}
				}
			} 
		}
	}		
	
	return 0;
}

